@@ -18,6 +18,8 @@ import fnmatch
 import shutil
 import pathlib
 import tempfile
+import io
+import shlex
 
 import urllib.parse
 import xml.etree.ElementTree
@@ -25,9 +27,19 @@ import xml.etree.ElementTree
 import b4
 
 from typing import Optional, Tuple
+from string import Template
 
 logger = b4.logger
 
+DEFAULT_MERGE_TEMPLATE = """Merge ${patch_or_series} "${seriestitle}"
+
+${authorname} <${authoremail}> says:
+
+${covermessage}
+
+Link: ${midurl}
+"""
+
 
 def make_am(msgs, cmdargs, msgid):
     config = b4.get_main_config()
@@ -35,7 +47,6 @@ def make_am(msgs, cmdargs, msgid):
     if outdir == '-':
         cmdargs.nocover = True
     wantver = cmdargs.wantver
-    wantname = cmdargs.wantname
     covertrailers = cmdargs.covertrailers
     count = len(msgs)
     logger.info('Analyzing %s messages in the thread', count)
@@ -93,54 +104,10 @@ def make_am(msgs, cmdargs, msgid):
                                     covertrailers=covertrailers, trailer_order=config['trailer-order'],
                                     addmysob=cmdargs.addmysob, addlink=cmdargs.addlink,
                                     linkmask=config['linkmask'], cherrypick=cherrypick,
-                                    copyccs=cmdargs.copyccs)
+                                    copyccs=cmdargs.copyccs, allowbadchars=cmdargs.allowbadchars)
     except KeyError:
         sys.exit(1)
 
-    if cmdargs.maildir or config.get('save-maildirs', 'no') == 'yes':
-        save_maildir = True
-        dftext = 'maildir'
-    else:
-        save_maildir = False
-        dftext = 'mbx'
-
-    if wantname:
-        slug = wantname
-        if wantname.find('.') > -1:
-            slug = '.'.join(wantname.split('.')[:-1])
-        gitbranch = slug
-    else:
-        slug = lser.get_slug(extended=True)
-        gitbranch = lser.get_slug(extended=False)
-
-    if outdir != '-':
-        am_filename = os.path.join(outdir, f'{slug}.{dftext}')
-        am_cover = os.path.join(outdir, f'{slug}.cover')
-
-        if os.path.exists(am_filename):
-            if os.path.isdir(am_filename):
-                shutil.rmtree(am_filename)
-            else:
-                os.unlink(am_filename)
-        if save_maildir:
-            d_new = os.path.join(am_filename, 'new')
-            pathlib.Path(d_new).mkdir(parents=True)
-            d_cur = os.path.join(am_filename, 'cur')
-            pathlib.Path(d_cur).mkdir(parents=True)
-            d_tmp = os.path.join(am_filename, 'tmp')
-            pathlib.Path(d_tmp).mkdir(parents=True)
-            for m_slug, msg in am_msgs:
-                with open(os.path.join(d_tmp, f'{m_slug}.eml'), 'wb') as mfh:
-                    mfh.write(msg.as_bytes(policy=b4.emlpolicy))
-                os.rename(os.path.join(d_tmp, f'{m_slug}.eml'), os.path.join(d_new, f'{m_slug}.eml'))
-        else:
-            with open(am_filename, 'wb') as fh:
-                b4.save_git_am_mbox([x[1] for x in am_msgs], fh)
-    else:
-        am_filename = None
-        am_cover = None
-        b4.save_git_am_mbox([x[1] for x in am_msgs], sys.stdout.buffer)
-
     logger.info('---')
 
     if cherrypick is None:
@@ -182,6 +149,17 @@ def make_am(msgs, cmdargs, msgid):
             logger.critical('     Msg From: %s <%s>', fname, femail)
         logger.critical('NOTE: Rerun with -S to apply them anyway')
 
+    top_msgid = None
+    first_body = None
+    for lmsg in lser.patches:
+        if lmsg is not None:
+            first_body = lmsg.body
+            top_msgid = lmsg.msgid
+            break
+    if top_msgid is None:
+        logger.critical('Could not find any patches in the series.')
+        return
+
     topdir = b4.git_get_toplevel()
 
     if cmdargs.threeway:
@@ -203,81 +181,232 @@ def make_am(msgs, cmdargs, msgid):
     if not lser.complete and not cmdargs.cherrypick:
         logger.critical('WARNING: Thread incomplete!')
 
-    if lser.has_cover and not cmdargs.nocover:
-        lser.save_cover(am_cover)
+    gitbranch = lser.get_slug(extended=False)
+    am_filename = None
 
-    top_msgid = None
-    first_body = None
-    for lmsg in lser.patches:
-        if lmsg is not None:
-            first_body = lmsg.body
-            top_msgid = lmsg.msgid
-            break
-    if top_msgid is None:
-        logger.critical('Could not find any patches in the series.')
-        return
+    if cmdargs.subcmd == 'am':
+        wantname = cmdargs.wantname
+        if cmdargs.maildir or config.get('save-maildirs', 'no') == 'yes':
+            save_maildir = True
+            dftext = 'maildir'
+        else:
+            save_maildir = False
+            dftext = 'mbx'
 
-    linkurl = config['linkmask'] % top_msgid
-    if cmdargs.quiltready:
-        q_dirname = os.path.join(outdir, f'{slug}.patches')
-        save_as_quilt(am_msgs, q_dirname)
-        logger.critical('Quilt: %s', q_dirname)
+        if wantname:
+            slug = wantname
+            if wantname.find('.') > -1:
+                slug = '.'.join(wantname.split('.')[:-1])
+            gitbranch = slug
+        else:
+            slug = lser.get_slug(extended=True)
 
-    logger.critical(' Link: %s', linkurl)
+        if outdir != '-':
+            am_filename = os.path.join(outdir, f'{slug}.{dftext}')
+            am_cover = os.path.join(outdir, f'{slug}.cover')
+
+            if os.path.exists(am_filename):
+                if os.path.isdir(am_filename):
+                    shutil.rmtree(am_filename)
+                else:
+                    os.unlink(am_filename)
+            if save_maildir:
+                b4.save_maildir(am_msgs, am_filename)
+            else:
+                with open(am_filename, 'w') as fh:
+                    b4.save_git_am_mbox(am_msgs, fh)
+        else:
+            am_cover = None
+            b4.save_git_am_mbox(am_msgs, sys.stdout)
+
+        if lser.has_cover and not cmdargs.nocover:
+            lser.save_cover(am_cover)
+
+        linkurl = config['linkmask'] % top_msgid
+        if cmdargs.quiltready:
+            q_dirname = os.path.join(outdir, f'{slug}.patches')
+            save_as_quilt(am_msgs, q_dirname)
+            logger.critical('Quilt: %s', q_dirname)
+
+        logger.critical(' Link: %s', linkurl)
 
     base_commit = None
-    matches = re.search(r'base-commit: .*?([0-9a-f]+)', first_body, re.MULTILINE)
+    matches = re.search(r'base-commit: .*?([\da-f]+)', first_body, re.MULTILINE)
     if matches:
         base_commit = matches.groups()[0]
     else:
         # Try a more relaxed search
-        matches = re.search(r'based on .*?([0-9a-f]{40})', first_body, re.MULTILINE)
+        matches = re.search(r'based on .*?([\da-f]{40})', first_body, re.MULTILINE)
         if matches:
             base_commit = matches.groups()[0]
 
-    if base_commit:
-        logger.critical(' Base: %s', base_commit)
+    if not base_commit and topdir and cmdargs.guessbase:
+        logger.critical(' Base: attempting to guess base-commit...')
+        try:
+            base_commit, nblobs, mismatches = lser.find_base(topdir, branches=cmdargs.guessbranch,
+                                                             maxdays=cmdargs.guessdays)
+            if mismatches == 0:
+                logger.critical(' Base: %s (exact match)', base_commit)
+            elif nblobs == mismatches:
+                logger.critical(' Base: failed to guess base')
+            else:
+                logger.critical(' Base: %s (best guess, %s/%s blobs matched)', base_commit,
+                                nblobs - mismatches, nblobs)
+        except IndexError:
+            logger.critical(' Base: failed to guess base')
+
+    if cmdargs.subcmd == 'shazam':
+        if not topdir:
+            logger.critical('Could not figure out where your git dir is, cannot shazam.')
+            sys.exit(1)
+        ifh = io.StringIO()
+        b4.save_git_am_mbox(am_msgs, ifh)
+        ambytes = ifh.getvalue().encode()
+        if not cmdargs.makefetchhead:
+            amflags = config.get('shazam-am-flags', '')
+            sp = shlex.shlex(amflags, posix=True)
+            sp.whitespace_split = True
+            amargs = list(sp)
+            ecode, out = b4.git_run_command(topdir, ['am'] + amargs, stdin=ambytes, logstderr=True)
+            logger.info(out.strip())
+            if ecode == 0:
+                thanks_record_am(lser, cherrypick=cherrypick)
+            sys.exit(ecode)
+
+        if not base_commit:
+            # Try our best with HEAD, I guess
+            base_commit = 'HEAD'
+
+        with b4.git_temp_worktree(topdir, base_commit) as gwt:
+            logger.info('Magic: Preparing a sparse worktree')
+            ecode, out = b4.git_run_command(gwt, ['sparse-checkout', 'init'], logstderr=True)
+            if ecode > 0:
+                logger.critical('Error running sparse-checkout init')
+                logger.critical(out)
+                sys.exit(ecode)
+            ecode, out = b4.git_run_command(gwt, ['checkout'], logstderr=True)
+            if ecode > 0:
+                logger.critical('Error running checkout into sparse workdir')
+                logger.critical(out)
+                sys.exit(ecode)
+            ecode, out = b4.git_run_command(gwt, ['am'], stdin=ambytes, logstderr=True)
+            if ecode > 0:
+                logger.critical('Unable to cleanly apply series, see failure log below')
+                logger.critical('---')
+                logger.critical(out.strip())
+                logger.critical('---')
+                logger.critical('Not fetching into FETCH_HEAD')
+                sys.exit(ecode)
+            logger.info('---')
+            logger.info(out.strip())
+            logger.info('---')
+            logger.info('Fetching into FETCH_HEAD')
+            gitargs = ['fetch', gwt]
+            ecode, out = b4.git_run_command(topdir, gitargs, logstderr=True)
+            if ecode > 0:
+                logger.critical('Unable to fetch from the worktree')
+                logger.critical(out.strip())
+                sys.exit(ecode)
+            gitargs = ['rev-parse', '--git-path', 'FETCH_HEAD']
+            ecode, fhf = b4.git_run_command(topdir, gitargs, logstderr=True)
+            if ecode > 0:
+                logger.critical('Unable to find FETCH_HEAD')
+                logger.critical(out.strip())
+                sys.exit(ecode)
+            with open(fhf.rstrip(), 'r') as fhh:
+                contents = fhh.read()
+            linkurl = config['linkmask'] % top_msgid
+            if len(am_msgs) > 1:
+                mmsg = 'patches from %s' % linkurl
+            else:
+                mmsg = 'patch from %s' % linkurl
+            new_contents = contents.replace(gwt, mmsg)
+            if new_contents != contents:
+                with open(fhf, 'w') as fhh:
+                    fhh.write(new_contents)
+
+            gitargs = ['rev-parse', '--git-dir']
+            ecode, fhf = b4.git_run_command(topdir, gitargs, logstderr=True)
+            if ecode > 0:
+                logger.critical('Unable to find git directory')
+                logger.critical(out.strip())
+                sys.exit(ecode)
+            mmf = os.path.join(fhf.rstrip(), 'b4-cover')
+            merge_template = DEFAULT_MERGE_TEMPLATE
+            if config.get('shazam-merge-template'):
+                # Try to load this template instead
+                try:
+                    merge_template = b4.read_template(config['shazam-merge-template'])
+                except FileNotFoundError:
+                    logger.critical('ERROR: shazam-merge-template says to use %s, but it does not exist',
+                                    config['shazam-merge-template'])
+                    sys.exit(2)
+
+            # Write out a sample merge message using the cover letter
+            if os.path.exists(mmf):
+                # Make sure any old cover letters don't confuse anyone
+                os.unlink(mmf)
+
+            if lser.has_cover:
+                cmsg = lser.patches[0]
+                parts = b4.LoreMessage.get_body_parts(cmsg.body)
+                covermessage = parts[1]
+            else:
+                cmsg = lser.patches[1]
+                covermessage = ('NOTE: No cover letter provided by the author.\n'
+                                '      Add merge commit message here.')
+            tptvals = {
+                'seriestitle': cmsg.subject,
+                'authorname': cmsg.fromname,
+                'authoremail': cmsg.fromemail,
+                'covermessage': covermessage,
+                'midurl': linkurl,
+            }
+            if len(am_msgs) > 1:
+                tptvals['patch_or_series'] = 'patch series'
+            else:
+                tptvals['patch_or_series'] = 'patch'
+
+            body = Template(merge_template).safe_substitute(tptvals)
+            with open(mmf, 'w') as mmh:
+                mmh.write(body)
+
+        mergeflags = config.get('shazam-merge-flags', '--signoff')
+        sp = shlex.shlex(mergeflags, posix=True)
+        sp.whitespace_split = True
+        mergeargs = ['merge', '--no-ff', '-F', mmf, '--edit', 'FETCH_HEAD'] + list(sp)
+        mergecmd = ['git'] + mergeargs
+
+        thanks_record_am(lser, cherrypick=cherrypick)
+        if cmdargs.merge:
+            if not cmdargs.no_interactive:
+                logger.info('Will exec: %s', ' '.join(mergecmd))
+                try:
+                    input('Press Enter to continue or Ctrl-C to abort')
+                except KeyboardInterrupt:
+                    logger.info('')
+                    sys.exit(130)
+            else:
+                logger.info('Invoking: %s', ' '.join(mergecmd))
+            # We exec git-merge and let it take over
+            os.execvp(mergecmd[0], mergecmd)
+
+        logger.info('You can now merge or checkout FETCH_HEAD')
+        logger.info('  e.g.: %s', ' '.join(mergecmd))
+        return
+
+    if not base_commit:
+        checked, mismatches = lser.check_applies_clean(topdir, at=cmdargs.guessbranch)
+        if checked and len(mismatches) == 0 and checked != mismatches:
+            logger.critical(' Base: applies clean to current tree')
+            base_commit = 'HEAD'
+        else:
+            logger.critical(' Base: not specified')
+
+    if base_commit is not None:
         logger.critical('       git checkout -b %s %s', gitbranch, base_commit)
-        if cmdargs.outdir != '-':
-            logger.critical('       git am %s', am_filename)
-    else:
-        cleanmsg = ''
-        if topdir is not None:
-            checked, mismatches = lser.check_applies_clean(topdir)
-            if mismatches == 0 and checked != mismatches:
-                cleanmsg = ' (applies clean to current tree)'
-            elif cmdargs.guessbase:
-                # Look at the last 10 tags and see if it applies cleanly to
-                # any of them. I'm not sure how useful this is, but I'm going
-                # to put it in for now and maybe remove later if it causes
-                # problems or slowness
-                if checked != mismatches:
-                    best_matches = mismatches
-                    cleanmsg = ' (best guess: current tree)'
-                else:
-                    best_matches = None
-                # sort the tags by authordate
-                gitargs = ['tag', '-l', '--sort=-taggerdate']
-                lines = b4.git_get_command_lines(None, gitargs)
-                if lines:
-                    # Check last 10 tags
-                    for tag in lines[:10]:
-                        logger.debug('Checking base-commit possibility for %s', tag)
-                        checked, mismatches = lser.check_applies_clean(topdir, tag)
-                        if mismatches == 0 and checked != mismatches:
-                            cleanmsg = ' (applies clean to: %s)' % tag
-                            break
-                        # did they all mismatch?
-                        if checked == mismatches:
-                            continue
-                        if best_matches is None or mismatches < best_matches:
-                            best_matches = mismatches
-                            cleanmsg = ' (best guess: %s)' % tag
-
-        logger.critical(' Base: not found%s', cleanmsg)
-        if cmdargs.outdir != '-':
-            logger.critical('       git am %s', am_filename)
-
+    if cmdargs.outdir != '-':
+        logger.critical('       git am %s', am_filename)
     thanks_record_am(lser, cherrypick=cherrypick)
 
 
@@ -350,37 +479,24 @@ def save_as_quilt(am_msgs, q_dirname):
         return
     pathlib.Path(q_dirname).mkdir(parents=True)
     patch_filenames = list()
-    with tempfile.TemporaryDirectory() as tfd:
-        m_out = os.path.join(tfd, 'm')
-        p_out = os.path.join(tfd, 'p')
-        for slug, msg in am_msgs:
-            # Run each message through git mailinfo
-            cmdargs = ['mailinfo', '--encoding=UTF-8', '--scissors', m_out, p_out]
-            ecode, info = b4.git_run_command(None, cmdargs, msg.as_bytes(policy=b4.emlpolicy))
-            if not len(info.strip()):
-                logger.critical('ERROR: Could not get mailinfo from patch %s', msg.get('Subject', '(no subject)'))
-                continue
-            patchinfo = dict()
-            for line in info.split('\n'):
-                line = line.strip()
-                if not line:
-                    continue
-                chunks = line.split(':',  1)
-                patchinfo[chunks[0]] = chunks[1].strip().encode()
-
-            patch_filename = f'{slug}.patch'
-            patch_filenames.append(patch_filename)
-            quilt_out = os.path.join(q_dirname, patch_filename)
-            with open(quilt_out, 'wb') as fh:
-                fh.write(b'From: %s <%s>\n' % (patchinfo['Author'], patchinfo['Email']))
-                fh.write(b'Subject: %s\n' % patchinfo['Subject'])
-                fh.write(b'Date: %s\n' % patchinfo['Date'])
-                fh.write(b'\n')
-                with open(m_out, 'rb') as mfh:
-                    shutil.copyfileobj(mfh, fh)
-                with open(p_out, 'rb') as pfh:
-                    shutil.copyfileobj(pfh, fh)
-            logger.debug('  Wrote: %s', patch_filename)
+    for msg in am_msgs:
+        lsubj = b4.LoreSubject(msg.get('subject', ''))
+        slug = '%04d_%s' % (lsubj.counter, re.sub(r'\W+', '_', lsubj.subject).strip('_').lower())
+        patch_filename = f'{slug}.patch'
+        patch_filenames.append(patch_filename)
+        quilt_out = os.path.join(q_dirname, patch_filename)
+        i, m, p = b4.get_mailinfo(msg.as_bytes(policy=b4.emlpolicy), scissors=True)
+        with open(quilt_out, 'wb') as fh:
+            if i.get('Author'):
+                fh.write(b'From: %s <%s>\n' % (i.get('Author').encode(), i.get('Email').encode()))
+            else:
+                fh.write(b'From: %s\n' % i.get('Email').encode())
+            fh.write(b'Subject: %s\n' % i.get('Subject').encode())
+            fh.write(b'Date: %s\n' % i.get('Date').encode())
+            fh.write(b'\n')
+            fh.write(m)
+            fh.write(p)
+        logger.debug('  Wrote: %s', patch_filename)
     # Write the series file
     with open(os.path.join(q_dirname, 'series'), 'w') as sfh:
         for patch_filename in patch_filenames:
@@ -399,7 +515,7 @@ def get_extra_series(msgs: list, direction: int = 1, wantvers: Optional[int] = N
         seen_msgids.add(msgid)
         lsub = b4.LoreSubject(msg['Subject'])
         if direction > 0 and lsub.reply:
-            # Does it have an "Obsoleted-by: trailer?
+            # Does it have an Obsoleted-by: trailer?
             rmsg = b4.LoreMessage(msg)
             trailers, mismatches = rmsg.get_trailers()
             for tl in trailers:
@@ -431,15 +547,14 @@ def get_extra_series(msgs: list, direction: int = 1, wantvers: Optional[int] = N
     config = b4.get_main_config()
     loc = urllib.parse.urlparse(config['midmask'])
     if not useproject:
-        projects = b4.get_lore_projects_from_msg(base_msg)
-        if not projects:
-            logger.info('Unable to figure out list archive location')
-            return msgs
-        useproject = projects[0]
+        useproject = 'all'
 
-    listarc = '%s://%s/%s/' % (loc.scheme, loc.netloc, useproject)
-
-    if not listarc:
+    listarc = f'{loc.scheme}://{loc.netloc}/{useproject}/'
+    # Make sure it exists
+    queryurl = f'{listarc}_/text/config/raw'
+    session = b4.get_requests_session()
+    resp = session.get(queryurl)
+    if not resp.status_code == 200:
         logger.info('Unable to figure out list archive location')
         return msgs
 
@@ -451,6 +566,7 @@ def get_extra_series(msgs: list, direction: int = 1, wantvers: Optional[int] = N
             t_mbx_url = '%s/%s/t.mbox.gz' % (listarc.rstrip('/'), nt_msgid)
             potentials = b4.get_pi_thread_by_url(t_mbx_url, nocache=nocache)
             if potentials:
+                potentials = b4.get_strict_thread(potentials, nt_msgid)
                 nt_msgs += potentials
                 logger.info('   Added %s messages from that thread', len(potentials))
             else:
@@ -494,6 +610,7 @@ def get_extra_series(msgs: list, direction: int = 1, wantvers: Optional[int] = N
         resp.close()
         ns = {'atom': 'http://www.w3.org/2005/Atom'}
         entries = tree.findall('atom:entry', ns)
+        seen_urls = set()
 
         for entry in entries:
             title = entry.find('atom:title', ns).text
@@ -530,6 +647,9 @@ def get_extra_series(msgs: list, direction: int = 1, wantvers: Optional[int] = N
                 logger.debug('No idea what this is: %s', title)
                 continue
             t_mbx_url = '%st.mbox.gz' % link
+            if t_mbx_url in seen_urls:
+                continue
+            seen_urls.add(t_mbx_url)
             logger.info('New revision: %s', title)
             potentials = b4.get_pi_thread_by_url(t_mbx_url, nocache=nocache)
             if potentials:
@@ -559,12 +679,9 @@ def get_msgs(cmdargs) -> Tuple[Optional[str], Optional[list]]:
             sys.exit(1)
 
         pickings = set()
-        try:
-            if cmdargs.cherrypick == '_':
-                # Just that msgid, please
-                pickings = {msgid}
-        except AttributeError:
-            pass
+        if 'cherrypick' in cmdargs and cmdargs.cherrypick == '_':
+            # Just that msgid, please
+            pickings = {msgid}
         msgs = b4.get_pi_thread_by_msgid(msgid, useproject=cmdargs.useproject, nocache=cmdargs.nocache,
                                          onlymsgids=pickings)
         if not msgs:
@@ -596,6 +713,9 @@ def get_msgs(cmdargs) -> Tuple[Optional[str], Optional[list]]:
             logger.critical('Mailbox %s does not exist', cmdargs.localmbox)
             sys.exit(1)
 
+    if msgid and 'noparent' in cmdargs and cmdargs.noparent:
+        msgs = b4.get_strict_thread(msgs, msgid, noparent=True)
+
     if not msgid and msgs:
         for msg in msgs:
             msgid = msg.get('Message-ID', None)
@@ -607,6 +727,20 @@ def get_msgs(cmdargs) -> Tuple[Optional[str], Optional[list]]:
 
 
 def main(cmdargs):
+    if cmdargs.subcmd == 'shazam':
+        # We force some settings
+        cmdargs.checknewer = True
+        cmdargs.threeway = False
+        cmdargs.nopartialreroll = False
+        cmdargs.outdir = '-'
+        cmdargs.guessbranch = None
+        if cmdargs.merge:
+            cmdargs.makefetchhead = True
+        if cmdargs.makefetchhead:
+            cmdargs.guessbase = True
+        else:
+            cmdargs.guessbase = False
+
     if cmdargs.checknewer:
         # Force nocache mode
         cmdargs.nocache = True
@@ -618,14 +752,14 @@ def main(cmdargs):
     if len(msgs) and cmdargs.checknewer:
         msgs = get_extra_series(msgs, direction=1, useproject=cmdargs.useproject)
 
-    if cmdargs.subcmd == 'am':
+    if cmdargs.subcmd in ('am', 'shazam'):
         make_am(msgs, cmdargs, msgid)
         return
 
     logger.info('%s messages in the thread', len(msgs))
     if cmdargs.outdir == '-':
         logger.info('---')
-        b4.save_git_am_mbox(msgs, sys.stdout.buffer)
+        b4.save_git_am_mbox(msgs, sys.stdout)
         return
 
     # Check if outdir is a maildir
@@ -669,7 +803,7 @@ def main(cmdargs):
         logger.info('Saved maildir %s', savename)
         return
 
-    with open(savename, 'wb') as fh:
+    with open(savename, 'w') as fh:
         b4.save_git_am_mbox(msgs, fh)
 
     logger.info('Saved %s', savename)
